# Copyright 2023 Luminary Cloud, Inc. All Rights Reserved.

import logging
from typing import Any, Iterable, Optional, Tuple, cast

import grpc

from .._auth import Auth0Client
from .._proto.mesh.mesh_pb2_grpc import MeshServiceStub
from .._proto.project.project_pb2_grpc import ProjectServiceStub
from .._proto.simulation.simulation_pb2_grpc import SimulationServiceStub
from .auth0_metadata_plugin import Auth0MetadataPlugin
from .config import LC_DOMAIN
from .logging_interceptor import LoggingInterceptor
from .retry_interceptor import RetryInterceptor
from .tracing import add_instrumentation

logger = logging.getLogger(__name__)


class Client(
    ProjectServiceStub,
    MeshServiceStub,
    SimulationServiceStub,
):
    """
    Creates a Luminary API client.

    The returned client automatically obtains access tokens for the Luminary API and
    sends them with each RPC call. See auth/auth.py for details.

    Supports "with" syntax.

    Parameters
    ----------
    target : str
        The URL of the API server.
    localhost : bool
        True if the API server is running locally.
    grpc_channel_options : Optional[Iterable[tuple[str, str]]]
        A list of gRPC channel args. The full list is available here:
        https://github.com/grpc/grpc/blob/v1.46.x/include/grpc/impl/codegen/grpc_types.h
    **kwargs : dict, optional
        Additional arguments are passed to Auth0Client. See auth/auth.py.
    """

    def __init__(
        self,
        target: str = LC_DOMAIN,
        localhost: bool = False,
        grpc_channel_options: Optional[Iterable[Tuple[str, str]]] = None,
        **kwargs: Any,
    ):
        self._channel = _create_channel(target, localhost, grpc_channel_options, **kwargs)
        ProjectServiceStub.__init__(self, self._channel)
        MeshServiceStub.__init__(self, self._channel)
        SimulationServiceStub.__init__(self, self._channel)

    @property
    def channel(self) -> grpc.ChannelCredentials:
        return self._channel

    def __enter__(self) -> "Client":
        return self

    def __exit__(self, *exc: Any) -> None:
        self._channel.close()


def _create_channel(
    target: str,
    localhost: bool = False,
    grpc_channel_options: Optional[Iterable[Tuple[str, str]]] = None,
    channel_credentials: Optional[grpc.ChannelCredentials] = None,
    **kwargs: Any,
) -> grpc.ChannelCredentials:
    if channel_credentials is None:
        if localhost:
            logger.debug("Using local channel credentials.")
            channel_credentials = grpc.local_channel_credentials()
        else:
            logger.debug("Using SSL channel credentials.")
            channel_credentials = grpc.ssl_channel_credentials()
    auth0_client = Auth0Client(**kwargs)
    metadata_plugin = Auth0MetadataPlugin(auth0_client)
    auth0_credentials = grpc.metadata_call_credentials(metadata_plugin)
    options = grpc_channel_options and list(grpc_channel_options)
    channel = grpc.secure_channel(
        target,
        grpc.composite_channel_credentials(
            cast(grpc.ChannelCredentials, channel_credentials),
            auth0_credentials,
        ),
        options=options,
    )
    intercepted_channel = grpc.intercept_channel(
        channel,
        LoggingInterceptor(),
        RetryInterceptor(),
    )
    return add_instrumentation(intercepted_channel, target, auth0_client)
