# Copyright 2023 Luminary Cloud, Inc. All Rights Reserved.
import errno
import logging
import os
from pathlib import Path
from typing import Iterator, Callable, Optional, Union, cast

from .file_chunk_stream import FileChunkStream
from .._proto.common import common_pb2 as commonpb
from .._proto.simulation.simulation_pb2 import (
    GetSimulationSurfaceSolutionRequest,
    GetSimulationVolumeSolutionRequest,
)
from .._client import Client

logger = logging.getLogger(__name__)


def _iter_chunks(
    fetch_chunk_with_offset: Callable[[int], commonpb.FileChunk],
) -> Iterator[commonpb.FileChunk]:
    offset = 0
    while True:
        chunk = fetch_chunk_with_offset(offset)
        yield chunk
        if chunk.bytes.last_chunk:
            return
        offset = chunk.bytes.next_offset


def _stream(
    fetch_chunk_with_offset: Callable[[int], commonpb.FileChunk],
) -> FileChunkStream:
    return FileChunkStream(_iter_chunks(fetch_chunk_with_offset))


def download_surface_solution(
    client: Client,
    simulation_id: str,
    time_step: Optional[int] = None,
) -> FileChunkStream:
    """
    Returns the download as a file-like object.

    The filename can be retrieved from the `filename` attribute of the returned object.

    Examples
    --------
    >>> with download_surface_solution(client, "your-simulation-id") as dl:
    ...     with open(dl.filename, "wb") as fp:
    ...         fp.write(dl.read())
    """

    def _fetch(offset: int) -> commonpb.FileChunk:
        request = GetSimulationSurfaceSolutionRequest(
            id=simulation_id,
            start_offset=offset,
        )
        if time_step:
            request.time_step = time_step
        response = client.GetSimulationSurfaceSolution(request)
        return response.file_chunk

    logger.info(f"Begin streaming surface data for simulation ID {simulation_id}.")
    return _stream(_fetch)


def download_volume_solution(
    client: Client,
    simulation_id: str,
) -> FileChunkStream:
    """
    Returns the download as a file-like object.

    The filename can be retrieved from the `filename` attribute of the returned object.

    Examples
    --------
    >>> with download_volume_solution(client, "your-simulation-id") as dl:
    ...     with open(dl.filename, "wb") as fp:
    ...         fp.write(dl.read())
    """

    def _fetch(offset: int) -> commonpb.FileChunk:
        request = GetSimulationVolumeSolutionRequest(
            id=simulation_id,
            start_offset=offset,
        )
        response = client.GetSimulationVolumeSolution(request)
        return response.file_chunk

    logger.info(f"Begin streaming volume solution for simulation ID {simulation_id}.")
    return _stream(_fetch)


def save_file(
    file_proto: commonpb.File,
    dest_dir: Union[os.PathLike, str],
    *,
    filename: Optional[str] = None,
) -> None:
    """
    Saves the file to the given directory.

    Parameters
    ----------
    file_proto: commonpb.File
    dest_dir: PathLike or str
        The destination directory.
    filename: str
        (Optional) The desired filename. Extension is appended if omitted.

    Raises
    ------
    FileNotFoundError
        If the directory does not exist.
    OSError
        If a file with the same name already exists.

    Examples
    --------
    >>> res = client.GetSimulationSurfaceQuantityOutput(req)
    >>> save_file(res.csv_file, "~/Downloads")
    """
    metadata = file_proto.metadata
    if filename is None:
        filename = metadata.name
    if "." not in filename:
        if len(metadata.ext) > 0:
            filename += "." + metadata.ext

    dest_path = Path(dest_dir) / cast(str, filename)
    if os.path.exists(dest_path):
        err = OSError(errno.EEXIST, "File already exists at location", str(dest_path))
        logger.error("File not found.", exc_info=err)
        raise err

    logger.info(f"Writing file to {dest_path}")
    with open(dest_path, "w+") as fp:
        fp.write(file_proto.data.decode())
